import re


def extract_srcset_from_image_tag(tag):
    return re.search(r'srcset="(?P<srcset>.*?)"', tag).group("srcset")


class FakeImage:
    def __init__(self, width, filter_spec, attrs):
        self.width = width
        self.filter_spec = filter_spec
        self.attrs = attrs



import types

from django import template
from django.conf import settings
from wagtail.images.templatetags.wagtailimages_tags import image

register = template.Library()


def monkeypatch_wagtail_as_syntax(image_node):
    """
    If you use wagtails 'as' syntax like this:

    {% srcset_image photo width-300 as thumbnail %}

    image_node["attrs"] are not resolved. This patch resolves
    srcset and attaches it to the returned rendition. It's a
    bit hacky but should work.
    """

    def render_patched(self, context):
        result = self._original_render(context)
        if self.output_var_name:
            rendition = context[self.output_var_name]
            rendition.srcset = self.attrs["srcset"].resolve(context)
            rendition.original = self.attrs["srcset"].original_url
        return result

    image_node._original_render = image_node.render
    image_node.render = types.MethodType(render_patched, image_node)
    return image_node


@register.tag(name="srcset_image")
def srcset_image(parser, token):
    image_node = image(parser, token)
    image_node.attrs["srcset"] = SrcSet(image_node)
    image_node = monkeypatch_wagtail_as_syntax(image_node)
    return image_node


class SrcSet:
    """
    Sets the srcset attribute of an image. You can either set it
    explicitly like this:

    {% srcset_image img width-600 srcset="width-1200 width-300" %}

    Or leave it out and let it be set automatically:

    {% srcset_image img width-600 %}

    Entries in the srcset attribute have the same syntax as normal
    wagtail image tag filter specs, although it wouldn't make much
    sence to use fill-80x80 or scale-50 inside srcset. At the moment
    width and jpegquality are the only supported operations.
    """

    ALLOWED_OPERATIONS = {"width", "jpegquality"}
    DEFAULT_WIDTHS = (2200, 1100, 768, 500, 300)
    DEFAULT_SRCSET_RENDITIONS = [f"width-{dw}" for dw in DEFAULT_WIDTHS]

    def __init__(self, image_node):
        self.image_node = image_node
        # need to set this inside __init__
        self.srcset_attribute = self.image_node.attrs.get("srcset", None)

    def get_allowed_operations(self, filter_spec):
        operations = []
        used_operations = set()
        for operation in filter_spec.split("|"):
            name = operation.split("-")[0]
            already_used = name in used_operations
            operation_allowed = name in self.ALLOWED_OPERATIONS
            if operation_allowed and not already_used:
                operations.append(operation)
                used_operations.add(name)
        return operations, used_operations

    def filter_specs_from_srcset(self, srcset):
        srcset = srcset.strip('"').strip("'")
        return srcset.split(" ")

    def merge_filter_specs(self, srcset_filter_specs, operations_from_tag):
        filter_specs = []
        for filter_spec in srcset_filter_specs:
            operations, used = self.get_allowed_operations(filter_spec)
            # inherit operations from tag if allowed and not already used
            # by the srcset attribute
            for operation in operations_from_tag:
                name = operation.split("-")[0]
                if name not in used:
                    operations.append(operation)
            filter_specs.append("|".join(operations))
        return filter_specs

    def get_default_scale_factors(self):
        max_width = max(self.DEFAULT_WIDTHS)
        return [dw / max_width for dw in self.DEFAULT_WIDTHS]

    def get_maximum_width(self, image):
        max_useful_width = image.width
        for operation in self.image_node.filter_spec.split("|"):
            name, rest = operation.split("-")
            if name == "width" and len(rest) > 0:
                # if width-300 is set, there's no point serving
                # a 2800px image - limit to 900px instead
                max_useful_width = int(rest) * 3
        return min(image.width, max_useful_width)

    def calc_filter_specs(self, image):
        max_width = self.get_maximum_width(image)
        filter_specs = []
        for scale in self.get_default_scale_factors():
            width = int(scale * max_width)
            filter_specs.append(f"width-{width}")
        return filter_specs

    def get_srcset_filter_specs(self, image):
        if self.srcset_attribute is not None:
            return self.filter_specs_from_srcset(self.srcset_attribute.token)
        if hasattr(settings, "DEFAULT_SRCSET_RENDITIONS"):
            return settings.DEFAULT_SRCSET_RENDITIONS
        if getattr(settings, "SRCSET_DYNAMIC", False):
            return self.calc_filter_specs(image)
        return self.DEFAULT_SRCSET_RENDITIONS

    def get_merged_filter_specs(self, image):
        srcset_filter_specs = self.get_srcset_filter_specs(image)
        operations_from_tag, _ = self.get_allowed_operations(
            self.image_node.filter_spec
        )
        return self.merge_filter_specs(srcset_filter_specs, operations_from_tag)

    def resolve(self, context):
        self.resolved_image = image = self.image_node.image_expr.resolve(context)
        out_renditions = []
        for filter_spec in self.get_merged_filter_specs(image):
            rendered_image = image.get_rendition(filter_spec)
            out_renditions.append(f"{rendered_image.url} {rendered_image.width}w")
        srcset_string = ", ".join(out_renditions)
        return srcset_string

    @property
    def original_url(self):
        return self.resolved_image.file.url



from django.views.generic import TemplateView
from wagtail.images.models import Image as WagtailImage


class MainView(TemplateView):
    template_name = "index.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["img"] = WagtailImage.objects.first()
        return context



import io
import os

import pytest
from django.core.files.uploadedfile import SimpleUploadedFile
from django.template import Context
from django.template.base import Template
from PIL import Image
from wagtail.images.models import Image as WagtailImage

from .utils import extract_srcset_from_image_tag


def create_small_rgb():
    # this is a small test jpeg
    img = Image.new("RGB", (200, 200), (255, 0, 0, 0))
    return img


@pytest.fixture()
def small_jpeg_io():
    rgb = create_small_rgb()
    im_io = io.BytesIO()
    rgb.save(im_io, format="JPEG", quality=60, optimize=True, progressive=True)
    im_io.seek(0)
    im_io.name = "testimage.jpg"
    return im_io


@pytest.fixture()
def small_uploaded_file(small_jpeg_io):
    simple_png = SimpleUploadedFile(
        name="test.png", content=small_jpeg_io.read(), content_type="image/png"
    )
    small_jpeg_io.seek(0)
    return simple_png


@pytest.fixture()
def image(small_uploaded_file):
    image = WagtailImage(file=small_uploaded_file)
    image.save()
    yield image
    # teardown
    os.unlink(image.file.path)


@pytest.fixture()
def srcset_template():
    return """
        {% load wagtail_srcset_tags %}
        {% srcset_image photo width-300 %}
    """


@pytest.fixture()
def rendered_result(srcset_template, image):
    t = Template(srcset_template)
    return t.render(Context({"photo": image}))


@pytest.fixture()
def rendered_srcset(rendered_result):
    return extract_srcset_from_image_tag(rendered_result)



import pytest
from django.template import Context
from django.template.base import Template

from wagtail_srcset.templatetags.wagtail_srcset_tags import SrcSet

from .utils import FakeImage, extract_srcset_from_image_tag


class TestWagtailImageTag:
    pytestmark = pytest.mark.django_db

    def test_default_wagtail_image_tag(self, image):
        template_text = """
            {% load wagtailimages_tags %}
            {% image photo width-300 %}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        assert "img alt" in result


class TestEmptySrcSetAttribute:
    pytestmark = pytest.mark.django_db

    def test_srcset_image_tag(self, rendered_result):
        assert "srcset" in rendered_result

    def test_default_renditions(self, rendered_srcset):
        for dw in SrcSet.DEFAULT_WIDTHS:
            assert str(dw) in rendered_srcset

    def test_settings_default_renditions(self, srcset_template, image, settings):
        t = Template(srcset_template)
        width = "537"
        settings.DEFAULT_SRCSET_RENDITIONS = [f"width-{width}"]
        result = t.render(Context({"photo": image}))
        srcset = extract_srcset_from_image_tag(result)
        assert width in srcset


class TestSrcSetImageTag:
    pytestmark = pytest.mark.django_db

    def test_srcset_image_with_alt(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 alt="some description" %}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        assert 'alt="some description"' in result

    def test_empty_srcset_inherits_jpegquality_from_tag(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 jpegquality-90%}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        srcset = extract_srcset_from_image_tag(result)
        assert "jpegquality-90" in srcset

    def test_srcset_operation_doesnt_get_overwritten(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 jpegquality-90 srcset="width-600|jpegquality-40" %}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        srcset = extract_srcset_from_image_tag(result)
        assert "jpegquality-40" in srcset
        assert "jpegquality-90" not in srcset

    def test_use_first_operation(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 jpegquality-90 jpegquality-80 %}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        srcset = extract_srcset_from_image_tag(result)
        assert "jpegquality-90" in srcset
        assert "jpegquality-80" not in srcset

    def test_srcset_use_first_operation(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 srcset="width-600|jpegquality-40|jpegquality-30" %}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        srcset = extract_srcset_from_image_tag(result)
        assert "jpegquality-40" in srcset
        assert "jpegquality-30" not in srcset


class TestSrcSetDynamicSizes:
    pytestmark = pytest.mark.django_db

    def test_srcset_tag_default_scales(self, settings):
        settings.SRCSET_DYNAMIC = True
        fake_image = FakeImage(900, "max-300x200", {})
        srcset = SrcSet(fake_image)
        filter_specs = srcset.get_srcset_filter_specs(fake_image)
        assert fake_image.width == int(filter_specs[0].split("-")[-1])

    def test_srcset_tag_max_image_width(self, settings):
        settings.SRCSET_DYNAMIC = True
        fake_image = FakeImage(900, "width-600", {})
        srcset = SrcSet(fake_image)
        filter_specs = srcset.get_srcset_filter_specs(fake_image)
        assert fake_image.width == int(filter_specs[0].split("-")[-1])

    def test_srcset_tag_max_3_times_tag_width(self, settings):
        settings.SRCSET_DYNAMIC = True
        tag_width = 600
        fake_image = FakeImage(2800, f"width-{tag_width}", {})
        srcset = SrcSet(fake_image)
        filter_specs = srcset.get_srcset_filter_specs(fake_image)
        assert tag_width * 3 == int(filter_specs[0].split("-")[-1])


class TestSrcSetImageTagAsSyntax:
    pytestmark = pytest.mark.django_db

    def test_rendition_has_srcset_attribute(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 as thumbnail %}
            {{ thumbnail.srcset }}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        assert "width-300" in result

    def test_rendition_has_original_attribute(self, image):
        template_text = """
            {% load wagtail_srcset_tags %}
            {% srcset_image photo width-300 as thumbnail %}
            {{ thumbnail.original }}
        """
        t = Template(template_text)
        result = t.render(Context({"photo": image}))
        assert "original_images" in result



from django.conf import settings
from django.conf.urls.static import static
from django.urls import path

from .views import MainView

urlpatterns = [
    path("", MainView.as_view(), name="main"),
] + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
